
import logging
import sys
import time
import pkgutil
import yaml
import re

from publicaddr import constants
from publicaddr.constants import *
from publicaddr import randprov
from publicaddr import handlers

cfg = {}

def lookup(providers=constants.ALL, retries=None, timeout=None, ip=None):
    """lookup for all public IP if existsfrom random providers"""
    global cfg
    if retries is not None: cfg["retries"] = retries
    if timeout is not None: cfg["timeout"] = timeout
    logging.debug("lookup from providers retries=%s" % (cfg["retries"]))

    addrs = {"ip4": None, "ip6": None}

    for _ in range(cfg["retries"]):
        start = time.perf_counter()

        if providers == constants.ALL:
            provider = randprov.pick_all()
        elif providers == constants.HTTPS:
            provider = randprov.pick_https()
        elif providers == constants.DNS:
            provider = randprov.pick_dns()
        elif providers == constants.STUN:
            provider = randprov.pick_stun()
        else:
            logging.error("fetch invalid providers - %s" % providers)
            return addrs

        if provider["mode"] == constants.HTTPS:
            insecure = False
            pattern = provider["pattern"] if "pattern" in provider else None

            if "insecure" in provider:
                insecure = provider["insecure"]

            if ip == constants.IPv4 or ip == None:
                addrs["ip4"] = handlers.lookup_http(url=provider["url"], ipversion=constants.IPv4,
                                                    timeout=cfg["timeout"], insecure=insecure,
                                                    pattern=pattern)
            if ip == constants.IPv6 or ip == None:
                addrs["ip6"] = handlers.lookup_http(url=provider["url"], ipversion=constants.IPv6,
                                                    timeout=cfg["timeout"], insecure=insecure,
                                                    pattern=pattern)

        if provider["mode"] == constants.DNS:
            dnsclass = provider["class"] if "class" in provider else "IN"
            qtype = provider["qtype"] if "qtype" in provider else None
            pattern = provider["pattern"] if "pattern" in provider else None

            if ip == constants.IPv4 or ip == None:
                addrs["ip4"] = handlers.lookup_dns_v4(nameservers=provider["nameservers"],
                                                    lookup=provider["lookup"],
                                                    dnsclass=dnsclass, qtype=qtype,
                                                    timeout=cfg["timeout"], pattern=pattern)
            if ip == constants.IPv6 or ip == None:
                addrs["ip6"] = handlers.lookup_dns_v6(nameservers=provider["nameservers"],
                                                    lookup=provider["lookup"],
                                                    dnsclass=dnsclass, qtype=qtype,
                                                    timeout=cfg["timeout"], pattern=pattern)

        if provider["mode"] == constants.STUN:
            if ip == constants.IPv4 or ip == None:
                addrs["ip4"] = handlers.lookup_stun(host=provider["host"], port=provider["port"], 
                                                    ipversion=constants.IPv4,
                                                    transport=provider["transport"],
                                                    timeout=cfg["timeout"])
            if ip == constants.IPv6 or ip == None:
                addrs["ip6"] = handlers.lookup_stun(host=provider["host"], port=provider["port"], 
                                                    ipversion=constants.IPv6,
                                                    transport=provider["transport"],
                                                    timeout=cfg["timeout"])

        addrs["provider"] = provider["name"]
        addrs["proto"] = provider["mode"]

        request_time = time.perf_counter() - start
        addrs["duration"] = "{0:.3f}".format(request_time)

        if not(addrs["ip4"] is None and addrs["ip6"] is None):
            break
    return addrs

def _get_provider(name, proto):
    global cfg

    if proto == constants.HTTPS:
        for prov in cfg["providers"][constants.HTTPS]:
            if prov["name"] == name:
                return prov
        return None

    elif proto == constants.DNS:
        for prov in cfg["providers"][constants.DNS]:
            if prov["name"] == name:
                return prov
        return None

    elif proto == constants.STUN:
        for prov in cfg["providers"][constants.STUN]:
            if prov["name"] == name:
                return prov
        return None

    else:
        logging.error("invalid proto - %s" % proto)
        return None

def get(provider="google", proto=constants.DNS, ip=constants.IPv4, timeout=None):
    global cfg
    if timeout is not None: cfg["timeout"] = timeout

    # search provider
    _provider = _get_provider(provider, proto) 
    if _provider is None: return None
        
    logging.debug("get from provider=%s ipversion=%s ipproto=%s" % (_provider["name"], ip, _provider["mode"]))

    start = time.perf_counter()
    addr = {}

    if _provider["mode"] == constants.HTTPS:
        insecure = _provider["insecure"] if "insecure" in _provider else False
        pattern = _provider["pattern"] if "pattern" in _provider else None
        addr["ip"] = handlers.lookup_http(url=_provider["url"], ipversion=ip,
                                          timeout=cfg["timeout"], insecure=insecure,
                                          pattern=pattern)

    if _provider["mode"] == constants.DNS:
        dnsclass = _provider["class"] if "class" in _provider else "IN"
        qtype = _provider["qtype"] if "qtype" in _provider else None
        pattern = _provider["pattern"] if "pattern" in _provider else None

        dns_handler = handlers.lookup_dns_v4
        if ip == constants.IPv6:
            dns_handler = handlers.lookup_dns_v6
        addr["ip"] = dns_handler(nameservers=_provider["nameservers"], lookup=_provider["lookup"],
                                 dnsclass=dnsclass, qtype=qtype, timeout=cfg["timeout"], pattern=pattern)

    if _provider["mode"] == constants.STUN:
        addr["ip"] = handlers.lookup_stun(host=_provider["host"], port=_provider["port"], ipversion=ip, 
                                          transport=_provider["transport"], timeout=cfg["timeout"])

    request_time = time.perf_counter() - start
    addr["duration"] = "{0:.3f}".format(request_time)
    return addr

def load_cfg():
    """read default config or from file"""
    cfg = None
    try:
        conf = pkgutil.get_data(__package__, 'publicaddr.yml')
        cfg =  yaml.safe_load(conf) 
    except Exception as e:
        logging.error("invalid default config: %s" % e)
    return cfg

def init():
    """load default providers"""
    global cfg
    _cfg = load_cfg()
    if _cfg is None: return
    cfg = _cfg

    if cfg["debug"]:
        loglevel = logging.DEBUG
    else:
        loglevel = logging.ERROR

    logging.getLogger("requests").setLevel(loglevel)
    logging.getLogger("urllib3").setLevel(loglevel)
    logging.basicConfig(format='%(levelname)s %(message)s', stream=sys.stdout, level=loglevel)

    # load default providers
    pattern = r"[^A-Za-z0-9]+"
    for proto, providers in cfg["providers"].items():

        for prov in providers:
            if prov["name"] not in globals():
                prov_name = re.sub(pattern, '_', prov["name"].upper())
                globals()[prov_name] = prov["name"]

        if proto == constants.HTTPS:
            randprov.set_http_providers(providers)
        if proto == constants.DNS:
            randprov.set_dns_providers(providers)
        if proto == constants.STUN:
            randprov.set_stun_providers(providers)

init()


import unittest
import publicaddr
class TestCloudflare(unittest.TestCase):
    def test_get_ip4_dns(self):
        """get dns ip4"""
        ip = publicaddr.get(provider=publicaddr.CLOUDFLARE, ip=publicaddr.IPv4)
        self.assertNotEqual(ip, "")
    def test_get_ip6_dns(self):
        """get dns ip6"""
        ip = publicaddr.get(provider=publicaddr.CLOUDFLARE, ip=publicaddr.IPv6)
        self.assertNotEqual(ip, "")
    def test_get_ip4_http(self):
        """get http ip4"""
        ip = publicaddr.get(provider=publicaddr.CLOUDFLARE, ip=publicaddr.IPv4, proto=publicaddr.HTTPS)
        self.assertNotEqual(ip, "")
    def test_get_ip6_http(self):
        """get http ip6"""
        ip = publicaddr.get(provider=publicaddr.CLOUDFLARE, ip=publicaddr.IPv6, proto=publicaddr.HTTPS)
        self.assertNotEqual(ip, "")




from dataclasses import dataclass
import requests
import logging
import asyncio
import aiostun
import re
import socket

import dns.resolver
import dns.exception

from publicaddr import constants

# Suppress only the single warning from urllib3 needed.
from urllib3.exceptions import InsecureRequestWarning

def lookup_http(url, ipversion, timeout, insecure, pattern):
    ip = None
    requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

    try:
        if ipversion == constants.IPv4:
            def allowed_gai_family():
                return socket.AF_INET
        else:
            def allowed_gai_family():
                return socket.AF_INET6


        requests.packages.urllib3.util.connection.allowed_gai_family = allowed_gai_family

        ip = requests.get(url, timeout=timeout, verify=not insecure).text.rstrip()
    
        if pattern is not None:
            match = re.search(pattern, ip)
            if match:
                return match.group(1)
    except requests.exceptions.RequestException as e:
        logging.error("[http] - %s" % e)
    return ip

async def _lookup_stun(host, port, family, transport, timeout):
    async with aiostun.Client(host=host, port=port, family=family, proto=transport, timeout=timeout) as stunc:
        mapped_addr = await stunc.get_mapped_address()
    return mapped_addr

def lookup_stun(host, port, ipversion, transport, timeout):
    ip = None
    try:
        if ipversion == constants.IPv4:
            family = aiostun.IP4
        elif ipversion == constants.IPv6:
            family = aiostun.IP6
        else:
            family = aiostun.IP4

        if transport.lower() == "udp":
            transport = aiostun.UDP
        elif transport.lower() == "tcp":
            transport = aiostun.TCP
        elif transport.lower() == "tls":
            transport = aiostun.TLS
        else:
            transport = aiostun.UDP

        mapped_addr = asyncio.run(_lookup_stun(host, port, family, transport, timeout))
        ip = mapped_addr["ip"]
    except Exception as e:
        logging.error("[stun] - %s" % e)
    return ip

def _resolv_addr(nameservers=[], qname="localhost", rdtype="A", rdclass="IN", timeout=5.0, pattern=None):
    rdtype = rdtype.lower()

    dnsresolv = dns.resolver.Resolver(configure=False)
    dnsresolv.nameservers = nameservers
    dnsresolv.timeout = timeout
    dnsresolv.lifetime = timeout
    
    # make dns resolution and read only the first response
    answers = dnsresolv.resolve(qname, rdtype, rdclass)
    if rdtype == "txt" and len(answers):
        text = []
        for line in answers[0].strings:
            text.append(line.decode())
        if pattern is not None:
            match = re.search(pattern, "\n".join(text))
            if match:
                return match.group(1)
        return "\n".join(text)

    if rdtype in [ "a", "aaaa" ] and len(answers):
        return answers[0].to_text()

    return None

def lookup_dns_v4(nameservers, lookup, qtype, dnsclass, timeout, pattern):
    _qtype = "A" if qtype is None else qtype

    # resolvers nameservers
    resolver = dns.resolver.Resolver()
    _nameservers = []
    for n in nameservers:
        answer = resolver.resolve(n, "A")
        for rdata in answer:
            _nameservers.append(rdata.to_text())

    return lookup_dns(_nameservers, lookup, _qtype, dnsclass, timeout ,pattern)

def lookup_dns_v6(nameservers, lookup, qtype, dnsclass, timeout, pattern):
    _qtype = "AAAA" if qtype is None else qtype

    # resolvers nameservers
    resolver = dns.resolver.Resolver()
    _nameservers = []
    for n in nameservers:
        answer = resolver.resolve(n, "AAAA")
        for rdata in answer:
            _nameservers.append(rdata.to_text())

    return lookup_dns(_nameservers, lookup, _qtype, dnsclass, timeout, pattern)

def lookup_dns(nameservers, lookup, qtype, dnsclass, timeout, pattern):
    ip = None
    try:
        ip = _resolv_addr(nameservers=nameservers, qname=lookup, 
                          rdtype=qtype, rdclass=dnsclass,
                          timeout=timeout, pattern=pattern)
    except dns.exception.DNSException as e:
        logging.error("[dns] - %s" % e)
    return ip



import unittest
import publicaddr

class TestOpenDNS(unittest.TestCase):
    def test_get_ip4(self):
        """get dns ip4"""
        ip = publicaddr.get(provider=publicaddr.OPENDNS, ip=publicaddr.IPv4)
        self.assertNotEqual(ip, "")
    def test_get_ip6(self):
        """get dns ip6"""
        ip = publicaddr.get(provider=publicaddr.OPENDNS, ip=publicaddr.IPv6)
        self.assertNotEqual(ip, "")


import unittest
import publicaddr

class TestLookup(unittest.TestCase):
    def test_lookup(self):
        """get dns ip6 and ip4"""
        ips = publicaddr.lookup()
        self.assertIn("ip4", ips)
        self.assertIn("ip6", ips)



from random import randrange
from publicaddr import constants

_https_providers = []
_dns_providers = []
_stun_providers = []

def set_http_providers(providers):
    global _https_providers
    _https_providers.extend(providers)
    for prov in _https_providers:
        prov["mode"] = constants.HTTPS

def set_dns_providers(providers):
    global _dns_providers
    _dns_providers.extend(providers)
    for prov in _dns_providers:
        prov["mode"] = constants.DNS

def set_stun_providers(providers):
    global _stun_providers
    _stun_providers.extend(providers)
    for prov in _stun_providers:
        prov["mode"] = constants.STUN

def pick_all():
    global _https_providers 

    # merge all providers
    allproviders = []
    allproviders.extend(_https_providers)
    allproviders.extend(_dns_providers)
    allproviders.extend(_stun_providers)

    id = randrange(len(allproviders))
    return allproviders[id]

def pick_https():
    global _https_providers 
    id = randrange(len(_https_providers))
    return _https_providers[id]

def pick_dns():
    global _dns_providers 
    id = randrange(len(_dns_providers))
    return _dns_providers[id]

def pick_stun():
    global _stun_providers 
    id = randrange(len(_stun_providers))
    return _stun_providers[id]


#!/usr/bin/python

import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

KEYWORDS = ('public ip dns http google akamai opendns cloudflare')

setuptools.setup(
    name="publicaddr",
    version="0.13.0",
    author="Denis MACHARD",
    author_email="d.machard@gmail.com",
    description="Getting your public IP v4 and v6",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/dmachard/python-publicaddr",
    packages=['publicaddr', 'tests'],
    include_package_data=True,
    platforms='any',
    keywords=KEYWORDS,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
    ],
    install_requires=[
        "requests",
        "dnspython",
        "pyyaml",
        "aiostun",
    ]
)